// starting code for Lab 9

/*
  This program tests different sorting algorithms on random integer
  arrays of different sizes.  The default array size is 30 elements.
  NOTE: all methods in Arraytools.java must be implemented.

  Usage: java Sort <sort_name> [array_size]

  where <sort_name> is one of the following keywords:
     bubble, selection, insertion, shell, merge, merge2,
     quick, quick2, count1, count2, count3

  To view a sample data array: java Sort showdata [array_size]

  Examples:

  java Sort bubble
     tests bubble sort on a 30-element random data array

  java Sort selection 200
     tests selection sort on a 200-element random data array
*/

public class Sort extends SortBase {

    public int[] generateTestCase(int size) {
        return Arraytools.randomArray(size, 0, 4*size);   // default
        //return Arraytools.randomArray(size, 0, 100);    // for testing countSort1
        //return Arraytools.randomArray(size, 0, 2*size); // for testing countSort2
    }

    //------------------------------------------------------------
    
    public void bubbleSort(int[] data) {
        for (int i = 0; i < data.length; i++) {
            for (int j = data.length-1; j > i; j--) {
                if (data[j-1] > data[j]) {
                    Arraytools.swap(data, j-1, j);
                }
            }
        }
    }

    //------------------------------------------------------------
    
    public void selectionSort(int[] data) {
        for (int i = 0; i < data.length; i++) {
            int minpos = i;
            for (int j = i+1; j < data.length; j++) {
                if (data[j] < data[minpos]) {
                    minpos = j;
                }
            }
            Arraytools.swap(data, i, minpos);
        }
            
    }

    //------------------------------------------------------------

    public void insertionSort(int[] data) {
        for (int i = 1; i < data.length; i++) {
            int j = i-1;  // j points to the left of i
            while (j >= 0 && data[j] > data[j+1]) {
                Arraytools.swap(data, j, j+1);
                j--;
            }
        }
    }

    //------------------------------------------------------------
    // shell sort is a generalization of insertion sort
    
    public void shellSort(int[] data) {
        for (int gap=data.length/2; gap > 0; gap /= 2) {
            if (gap % 2 == 0) gap++;  // gap size must be odd
            for (int i = gap; i < data.length; i++) {
                int j = i-gap;  // j points to the left of i
                while (j >= 0 && data[j] > data[j+gap]) {
                    Arraytools.swap(data, j, j+gap);
                    j -= gap;
                }
            }
        }
    }
    //------------------------------------------------------------
    
    public void mergeSort(int[] data) {
        if (data.length <= 1) return;
        int[] firstHalf = new int[data.length/2];
        int[] secondHalf = new int[data.length - firstHalf.length];
        Arraytools.copy(data, 0, firstHalf.length-1, firstHalf, 0);
        Arraytools.copy(data, firstHalf.length, data.length-1, secondHalf, 0);
        mergeSort(firstHalf);
        mergeSort(secondHalf);
        mergeArrays(firstHalf, secondHalf, data);
    }

    // merges source1 and source2 arrays (sorted) into dest array
    public void mergeArrays(int[] source1, int[] source2, int[] dest) {
        int i = 0;
        int j = 0;
        int k = 0;
        while (i < source1.length && j < source2.length) {
            if (source1[i] <= source2[j]) {
                dest[k] = source1[i];
                i++;
            } else {
                dest[k] = source2[j];
                j++;
            }
            k++;
        }
        if (i < source1.length) {
            // copy remaining elements from source1
            Arraytools.copy(source1, i, source1.length-1, dest, k);
        } else {
            // copy remaining elements from source2
            Arraytools.copy(source2, j, source2.length-1, dest, k);
        }
    }

    //------------------------------------------------------------

    public void quickSort(int[] data) {
        quickSort(data, 0, data.length-1);
    }
    
    // recursive helper method that does most of the work
    public void quickSort(int[] data, int first, int last) {
        if (first >= last) return;
        int pivot = choosePivot(data, first, last);
        int mid = partition(data, first, last, pivot);
        quickSort(data, first, mid);
        quickSort(data, mid+1, last);
    }
    
    public int choosePivot(int[] data, int first, int last) {
        return -1;  // temporary, to allow compilation
    }

    public int partition(int[] data, int first, int last, int pivot) {
        /*
        initialize left index i to array position first
        initialize right index j to array position last
        loop:
            move left index rightward until we encounter an element >= pivot
            move right index leftward until we encounter an element <= pivot
            if indices have met or crossed, quit the loop
            otherwise swap elements at positions i and j
            move i to right and j to left one position, and continue loop
        return j as the midpoint position of the partition
        */

        return -1;  // temporary, to allow compilation
    }

    //------------------------------------------------------------
    // these are for later

    public void mergeSort2(int[] data) {}
    public void quickSort2(int[] data) {}
    public void countSort1(int[] data) {}
    public void countSort2(int[] data) {}
    public void countSort3(int[] data) {}

    //------------------------------------------------------------
}
