// SortBase cannot be run by itself, it must be extended by the Sort class.

import java.io.*;

abstract class SortBase {
    
    private String[] sortNames = {
        "bubble", "selection", "insertion", "shell", "merge", "merge2",
        "quick", "quick2", "count1", "count2", "count3"
    };

    public boolean validSortName(String x) {
        for (String n : sortNames) {
            if (x.equals(n)) return true;
        }
        return false;
    }
    
    public void run(String[] args) throws IOException {
        if (args.length == 0 || !args[0].equals("showdata") && !validSortName(args[0])) {
            System.out.println("Usage: java Sort <name> [array_size]");
            System.out.print("names: showdata ");
            for (String n : sortNames) System.out.printf("%s ", n);
            System.out.println();
            return;
        }
        String name = args[0];
        int size = 30;
        if (args.length == 2) size = Integer.parseInt(args[1]);
        int[] data = generateTestCase(size);
        System.out.println("Initial array:");
        Arraytools.print(data);
        if (name.equals("showdata")) {
            return;
        }
        int[] data2 = Arraytools.clone(data);
        System.out.println("Sorted array:");
        try {
            if (name.equals("bubble")) bubbleSort(data);
            if (name.equals("selection")) selectionSort(data);
            if (name.equals("insertion")) insertionSort(data);
            if (name.equals("shell")) shellSort(data);
            if (name.equals("merge")) mergeSort(data);
            if (name.equals("merge2")) mergeSort2(data);
            if (name.equals("quick")) quickSort(data);
            if (name.equals("quick2")) quickSort2(data);
            if (name.equals("count1")) countSort1(data);
            if (name.equals("count2")) countSort2(data);
            if (name.equals("count3")) countSort3(data);
        } catch (RuntimeException e) {
            System.out.printf("*** ERROR: %s\n", e);
        }
        Arraytools.print(data);
        // check if array was sorted correctly
        if (!Arraytools.isSorted(data) || !Arraytools.hasSameElements(data, data2)) {
            System.out.println("*** ERROR: array was not correctly sorted");
        } else {
            System.out.println("Success!");
        }
    }
    
    public void run_timing_tests(String[] args) throws IOException {
        if (args.length < 2 || !validSortName(args[0])) {
            System.out.println("Usage: java SortTimer <name> <start_size> [end_size] [increment] [filename]");
            System.out.print("names: ");
            for (String n : sortNames) System.out.printf("%s ", n);
            System.out.println();
            return;
        } 
        String name = args[0];
        PrintWriter out = null;
        String filename = "";
        // check if last command-line argument is a filename
        int numArgs;
        try {
            Integer.parseInt(args[args.length-1]);
            numArgs = args.length;
        } catch (NumberFormatException e) {
            filename = args[args.length-1];
            out = new PrintWriter(filename);
            numArgs = args.length-1;
        }
        if (numArgs == 2) {
            int start = Integer.parseInt(args[1]);
            time(name, start, start, 1000, out, filename);
        } else if (numArgs == 3) {
            int start = Integer.parseInt(args[1]);
            int end = Integer.parseInt(args[2]);
            time(name, start, end, 1000, out, filename);
        } else {
            int start = Integer.parseInt(args[1]);
            int end = Integer.parseInt(args[2]);
            int increment = Integer.parseInt(args[3]);
            time(name, start, end, increment, out, filename);
        }
    }

    public void time(String name, int start, int end, int increment,
                            PrintWriter out, String filename) {
        if (out != null) {
            out.printf("# Timing data for %s sort from size %d to %d by %d\n",
                       name, start, end, increment);
        }
        Stopwatch watch = new Stopwatch();
        for (int size = start; size <= end; size += increment) {
            int[] data = generateTestCase(size);
            int[] data2 = null;
            if (size == start) {
                data2 = Arraytools.clone(data);
            } else {
                data2 = null;
            }
            watch.start();
            if (name.equals("bubble")) bubbleSort(data);
            if (name.equals("selection")) selectionSort(data);
            if (name.equals("insertion")) insertionSort(data);
            if (name.equals("shell")) shellSort(data);
            if (name.equals("merge")) mergeSort(data);
            if (name.equals("merge2")) mergeSort2(data);
            if (name.equals("quick")) quickSort(data);
            if (name.equals("quick2")) quickSort2(data);
            if (name.equals("count1")) countSort1(data);
            if (name.equals("count2")) countSort2(data);
            if (name.equals("count3")) countSort3(data);
            watch.stop();
            if (size == start) {
                // verify (just once) that array was sorted correctly
                if (!Arraytools.isSorted(data) || !Arraytools.hasSameElements(data, data2)) {
                    throw new RuntimeException("*** ERROR: array was not correctly sorted");
                }
            }
            System.out.printf("%d \t%d\n", size, watch.getElapsedTime());
            if (out != null) {
                out.printf("%d \t%d\n", size, watch.getElapsedTime());
            }
        }
        if (out != null) {
            out.close();
            System.out.printf("Timing data saved in file \"%s\"\n", filename);
        }
        System.out.println("done");
    }

    //------------------------------------------------------------
    // don't change this method

    public static void main(String[] args) throws java.io.IOException {
        new Sort().run(args);
    }

    //------------------------------------------------------------
    // these are implemented in Sort.java

    abstract int[] generateTestCase(int size);
    abstract void bubbleSort(int[] data);
    abstract void selectionSort(int[] data);
    abstract void insertionSort(int[] data);
    abstract void shellSort(int[] data);
    abstract void mergeSort(int[] data);
    abstract void mergeSort2(int[] data);
    abstract void quickSort(int[] data);
    abstract void quickSort2(int[] data);
    abstract void countSort1(int[] data);
    abstract void countSort2(int[] data);
    abstract void countSort3(int[] data);
}
